/*
 * 
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 * 
 * Copyright 2007-2008 Sun Microsystems, Inc. All rights reserved.
 * 
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common Development
 * and Distribution License("CDDL") (collectively, the "License").  You
 * may not use this file except in compliance with the License. You can obtain
 * a copy of the License at https://glassfish.dev.java.net/public/CDDL+GPL.html
 * or glassfish/bootstrap/legal/LICENSE.txt.  See the License for the specific
 * language governing permissions and limitations under the License.
 * 
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at glassfish/bootstrap/legal/LICENSE.txt.
 * Sun designates this particular file as subject to the "Classpath" exception
 * as provided by Sun in the GPL Version 2 section of the License file that
 * accompanied this code.  If applicable, add the following below the License
 * Header, with the fields enclosed by brackets [] replaced by your own
 * identifying information: "Portions Copyrighted [year]
 * [name of copyright owner]"
 * 
 * Contributor(s):
 * 
 * If you wish your version of this file to be governed by only the CDDL or
 * only the GPL Version 2, indicate your decision by adding "[Contributor]
 * elects to include this software in this distribution under the [CDDL or GPL
 * Version 2] license."  If you don't indicate a single choice of license, a
 * recipient has the option to distribute your version of this file under
 * either the CDDL, the GPL Version 2 or to extend the choice of license to
 * its licensees as provided above.  However, if you add GPL Version 2 code
 * and therefore, elected the GPL Version 2 license, then the option applies
 * only if the new code is made subject to such option by the copyright
 * holder.
 *
 */
/*
 * BEGIN_HEADER - DO NOT EDIT
 * 
 * The contents of this file are subject to the terms
 * of the Common Development and Distribution License
 * (the "License").  You may not use this file except
 * in compliance with the License.
 *
 * You can obtain a copy of the license at
 * https://open-jbi-components.dev.java.net/public/CDDLv1.0.html.
 * See the License for the specific language governing
 * permissions and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL
 * HEADER in each file and include the License file at
 * https://open-jbi-components.dev.java.net/public/CDDLv1.0.html.
 * If applicable add the following below this CDDL HEADER,
 * with the fields enclosed by brackets "[]" replaced with
 * your own identifying information: Portions Copyright
 * [year] [name of copyright owner]
 */



package com.sun.grizzly.pipeline;

import java.nio.channels.SelectionKey;
import java.nio.ByteBuffer;
import java.nio.BufferUnderflowException;
import java.nio.channels.SocketChannel;

import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import java.util.logging.Level;
import java.util.Iterator;

import com.sun.enterprise.web.connector.grizzly.LinkedListPipeline;
import com.sun.enterprise.web.connector.grizzly.SelectorThread;
import com.sun.enterprise.web.connector.grizzly.Task;
import com.sun.enterprise.web.connector.grizzly.ReadTask;
import com.sun.enterprise.web.connector.grizzly.HtmlHelper;
import com.sun.enterprise.web.connector.grizzly.AsyncReadTask;

/**
 *
 * Throttle pipeline for throttling requests, per countext root.
 * 
 * @author James Tran
 * @author Jeanfrancois Arcand
 */
public class RequestThrottlePipeline extends LinkedListPipeline{
    
    private SelectorThread st;

    private final static String QUERY_STRING="?";   
    private final static String PATH_STRING="/";   
    
    private ExecutorService resumeThreadPool;

    ConcurrentHashMap<String, ConcurrentLinkedQueue<SelectionKeyByteBuffer>> 
            suspendedKeys = new ConcurrentHashMap<String, ConcurrentLinkedQueue<SelectionKeyByteBuffer>>();

    protected ConcurrentLinkedQueue<ByteBuffer> bbPool =
            new ConcurrentLinkedQueue<ByteBuffer>();
    
    
    private ThrottlePolicyManager throttlePolicyManager;

    public RequestThrottlePipeline () {
        super();
        this.throttlePolicyManager = new DefaultThrottlePolicyManager(this);
    }

    
    @Override
    public synchronized void addTask(Task task) {
        if (SelectorThread.logger().isLoggable(Level.FINER)) {
            SelectorThread.logger().log (Level.FINER,
                    "addTask called with Task [" + task + "]");
        }
        
        if (task.getType() == Task.READ_TASK) {            
            String requestURI = null;
            ByteBuffer bb = getReadByteBuffer();
            try {
                requestURI = peekRequestURI((ReadTask)task, bb);
            } catch (Throwable t) {
                AsyncReadTask asyncRT = (AsyncReadTask)task;
                asyncRT.terminate(false);
                returnReadByteBuffer(bb);            
                if (SelectorThread.logger().isLoggable(Level.FINER)) {
                    SelectorThread.logger().log (Level.FINER, "Error thrown by peekRequestURI - Task [" 
                            + task + "] : " + t.getLocalizedMessage(), t);
                }
                return;
            }

            
            if (requestURI != null) {
                if (SelectorThread.logger().isLoggable(Level.FINER)) {
                    SelectorThread.logger().log (Level.FINER, "Request URI is '" 
                            + requestURI + "'");
                }               
                ThrottlePolicy throttlePolicy 
                        = throttlePolicyManager.getThrottlePolicy(requestURI, port);
                
                 // Only associate throttler controler to throttlePolicy if throttlePolicy is configured for throttling
                if (throttlePolicy != null && throttlePolicy.getMaxConcurrentRequests() > 0) {
                    int maxConcurrencyLimit = throttlePolicy.getMaxConcurrentRequests();                    
                    int currentRequestCount = throttlePolicy.currentRequestCount();
                    
                    if (SelectorThread.logger().isLoggable(Level.FINER)) {
                        SelectorThread.logger().log(Level.FINER, 
                                "Found ThrottlePolicy for request URI '"
                                + requestURI + "' and port '" + port 
                                + "'; maxConcurrencyLimit="+maxConcurrencyLimit
                                +", currentRequestCount="+currentRequestCount);
                    }

                    String key = requestURI+":"+port;
                    // suspend selection key if maxed out concurrent messages
                    if (maxConcurrencyLimit != -1 && (currentRequestCount >= maxConcurrencyLimit)) {
                        // Disable keep-alive
                        task.getSelectionKey().attach(null);
                        st = task.getSelectorThread();
                        if (!suspendedKeys.containsKey(key)) {
                            suspendedKeys.put(key, new ConcurrentLinkedQueue<SelectionKeyByteBuffer>());
                        }
                        ConcurrentLinkedQueue<SelectionKeyByteBuffer> linkQ = suspendedKeys.get(key);
                        SelectionKey sk = task.getSelectionKey();
                        SelectionKeyByteBuffer skbb = new SelectionKeyByteBuffer(sk, bb);
                        linkQ.offer(skbb);
                        task.recycle();
                        st.returnTask(task);
                        if (SelectorThread.logger().isLoggable(Level.FINER)) {
                            SelectorThread.logger().log(Level.FINER, "Request URI '"
                                    + requestURI + "', port '" + port 
                                    + "'; SUSPENDED Selectionkey [" + sk + 
                                    "], ByteBuffer [" + bb + "] from Task [" 
                                    + task + "]; suspended size="+linkQ.size());
                        }     
                        throttlePolicy.incrementSuspendedCount();
                        throttlePolicyManager.applyThrottlePolicy(throttlePolicy);
                        return;                        
                    }
                }

                // Proceed with the read task - copying the contents of the "temp" ByteBuffer to
                // the one in ReadTask.
                ((ReadTask)task).getByteBuffer().put(bb);
                if (SelectorThread.logger().isLoggable(Level.FINER)) {
                    SelectorThread.logger().log(Level.FINER, "ReadTask ByteBuffer ["
                            + ((ReadTask)task).getByteBuffer()+ "]");
                }
                ((ReadTask)task).setBytesAvailable(true);

                if (throttlePolicy != null) {
                    throttlePolicy.incrementCurrentRequestCount();
                }
            }
            
            returnReadByteBuffer(bb);
        }
        super.addTask(task);
    }
    
    
    public void resumeSuspendedRequests(ThrottlePolicy throttlePolicy) {        
        if (resumeThreadPool == null) {
            resumeThreadPool = Executors.newCachedThreadPool();
        }
        resumeThreadPool.submit(new ResumerTask(this,throttlePolicy));        
    }    

    
    public synchronized void cleanup() {        
        if (resumeThreadPool != null) {
            resumeThreadPool.shutdown();
            if (SelectorThread.logger().isLoggable(Level.FINER)) {
                SelectorThread.logger().log(Level.FINER, "shutdown resume task thread pool");
            }
        }
        
        // Cancel any suspended tasks
        if (suspendedKeys != null && suspendedKeys.size() > 0) {
            for (Iterator<String> iter1=suspendedKeys.keySet().iterator(); iter1.hasNext();) {
                String uriPort = iter1.next();
                String uri = uriPort.substring(0,uriPort.indexOf(":"));
                String aport = uriPort.substring(uriPort.indexOf(":")+1, uriPort.length());
                ConcurrentLinkedQueue<SelectionKeyByteBuffer> linkQ = suspendedKeys.remove(uri);
                if (linkQ != null && linkQ.size() > 0) {
                    for (Iterator iter2=linkQ.iterator();iter2.hasNext();) {
                        SelectionKeyByteBuffer ssbb = linkQ.poll();
                        SelectionKey selkey = ssbb.getSelectionKey();
                        try {
                            Task task = st.getReadTask(selkey);
                            task.cancelTask("No resources available.", HtmlHelper.OK);
                        } catch (Throwable t) {
                            SelectorThread.logger().log(Level.WARNING,
                                    "Exception_during_cancel_suspended_selectionkey", t);
                        }
                    }
                }
                if (SelectorThread.logger().isLoggable(Level.FINER)) {
                    SelectorThread.logger().log(Level.FINER, 
                            "cancelled all suspended selection key read tasks for URI '"
                            + uri + "' on port '" + aport + "'");
                }        
            }
        }        
    }

    
    /***
     * Get the request URI from the {@link ByteBuffer}
     */
    protected String peekRequestURI(ReadTask readTask, ByteBuffer byteBuffer) throws Exception {
        SocketChannel socketChannel = 
                    (SocketChannel)readTask.getSelectionKey().channel();
        String token = null;
        if (socketChannel.isOpen()) {
            socketChannel.read(byteBuffer);
            int limit = byteBuffer.limit();
            int position = byteBuffer.position();
			
            if (parse (byteBuffer)) {
                byte[] chars = new byte[byteBuffer.limit() - byteBuffer.position()];

                byteBuffer.get(chars);

                token = new String(chars);

                int index = token.indexOf(0x20);
                if ( index != -1){
                    token = token.substring(0,index);
                }

                // Remove query string.
                index = token.indexOf(QUERY_STRING);
                if ( index != -1){
                    token = token.substring(0,index);
                }

                boolean slash = token.endsWith(PATH_STRING);
                if ( slash ){
                    token = token.substring(0,token.length() -1);
                }  
            } 

            byteBuffer.limit(limit);
            byteBuffer.position(position);
            byteBuffer.flip();
        }

        return token;
    }    

    protected ByteBuffer getReadByteBuffer() {
        int size = 1028;  // keep it small
        ByteBuffer bb = bbPool.poll();
        if (bb == null) {
            bb = ByteBuffer.allocate(size);
        }
        return bb;                
    }
    
    
    protected void returnReadByteBuffer(ByteBuffer bb) {
        bb.clear();
        bbPool.offer(bb);
    }
    
    
    protected void internalAddTask (Task task) {
        super.addTask(task);
    }
    
    /**
     * Parse the request line in search of the context-root bytes of the HTTP
     * Method. The {@link ByteBuffer} position and limit refer 
     * respectively to the start and the end of the context root.
     * @param byteBuffer The byteBuffer containing the requests bytes
     * @return true if the context-root has been found.
     */
    public boolean parse(ByteBuffer byteBuffer) {
        int curPosition = byteBuffer.position();
        int curLimit = byteBuffer.limit();
      
        // Rule a - If nothing, return to the Selector.
        if (byteBuffer.position() == 0)
            return false;
       
        byteBuffer.position(0);
        byteBuffer.limit(curPosition);
        int state =0;
        int start =0;
        int end = 0;        
        
        try {                         
            byte c;            
            
            // Rule b - try to determine the context-root
            while(byteBuffer.hasRemaining()) {
                c = byteBuffer.get();

                // State Machine
                // 0 - Search for the first SPACE ' ' between the method and the
                //     the request URI
                // 1 - Search for the second SPACE ' ' between the request URI
                //     and the method
                switch(state) {
                    case 0: // Search for first ' '
                        if (c == 0x20){
                            state = 1;
                            start = byteBuffer.position();
                            //start = byteBuffer.position() + 1;
                        }
                        break;
                    case 1: // Search for next ' '
                        if (c == 0x20){
                            end = byteBuffer.position() - 1;
                            return true;
                        }
                        break;
                    default:
                        throw new IllegalArgumentException("Unexpected state");
                }      
            }
            return false;
        } catch (BufferUnderflowException bue) {
            return false;
        } finally {     
            if ( end > 0 ){
                byteBuffer.position(start);
                byteBuffer.limit(end);
            } else {
                byteBuffer.limit(curLimit);
                byteBuffer.position(curPosition);                               
            }
        }       
    }
    
    
    class SelectionKeyByteBuffer {
        private SelectionKey sk;
        private ByteBuffer bb;
        
        SelectionKeyByteBuffer (SelectionKey sk, ByteBuffer bb) {
            this.sk = sk;
            this.bb = bb;
        }
        
        SelectionKey getSelectionKey() { return sk; }
        ByteBuffer getByteBuffer() {return bb; }
    }
    
    
    class ResumerTask implements Runnable {
        RequestThrottlePipeline pipeline;
        ThrottlePolicy throttlePolicy;
        ResumerTask (RequestThrottlePipeline pipeline, ThrottlePolicy throttlePolicy) {
            this.pipeline = pipeline;
            this.throttlePolicy = throttlePolicy;
        }

        public void run() {
            synchronized (pipeline) {
                String uri = throttlePolicy.getUrlContext();
                int port = throttlePolicy.getUrlPort();
                int resumeCount = throttlePolicy.getMaxConcurrentRequests() - throttlePolicy.currentRequestCount();
                String key = uri+":"+port;
                ConcurrentLinkedQueue<SelectionKeyByteBuffer> linkQ = pipeline.suspendedKeys.get(key);
                if (SelectorThread.logger().isLoggable(Level.FINER)) {
                    SelectorThread.logger().log(Level.FINER, "Resume SelectionKey(s) triggered; URI '"+uri+
                                             "', port '"+port+
                                             "'; resumeCount= "+resumeCount+
                                             ", suspended size="+linkQ.size());
                }
                if (linkQ != null && linkQ.size() > 0) {
                    for (int i=0; i < resumeCount; i++){
                        // only resume if there are still suspended ones..                        
                        if (linkQ.size()>0) {
                            SelectionKeyByteBuffer skbb = linkQ.poll();
                            try {
                                SelectionKey sk = skbb.getSelectionKey();
                                ByteBuffer bb = skbb.getByteBuffer();
                                if (SelectorThread.logger().isLoggable(Level.FINER)) {
                                    SelectorThread.logger().log(Level.FINER, 
                                            "Request URI '" + uri + "', port '"
                                            + port + "'; got cached Selectionkey [" 
                                            + sk + "] and cached ByteBuffer [" + bb + "]");
                                }        
                                AsyncReadTask task = (AsyncReadTask)st.getReadTask(sk);
                                task.getByteBuffer().put(bb);
                                task.setBytesAvailable(true);
                                throttlePolicy.incrementCurrentRequestCount();
                                pipeline.internalAddTask(task);
                                if (SelectorThread.logger().isLoggable(Level.FINER)) {
                                    SelectorThread.logger().log(Level.FINER, "Request URI '" 
                                            + uri + "', port '" + port 
                                            + "'; RESUMED Selectionkey [" 
                                            + task.getSelectionKey() + "],  ByteBuffer [" 
                                            + task.getByteBuffer() + "] on Task ["
                                            + task + "]");
                                }        
                                returnReadByteBuffer(bb);
                            } catch (Throwable t) {
                               SelectorThread.logger().log(Level.WARNING, 
                                       "Exception_during_resume_suspended_selectionkey", t);
                            }
                        }
                    }

                    if (linkQ.size() == 0) {
                        pipeline.suspendedKeys.remove(key);
                    }
                }                
            }
        }
    }

    public

    ThrottlePolicyManager getThrottlePolicyManager() {
        return throttlePolicyManager;
    }

    public void setThrottlePolicyManager(ThrottlePolicyManager throttlePolicyManager) {
        this.throttlePolicyManager = throttlePolicyManager;
    }
}

